import asyncio
import random

from RetroGeo import GeoLocator


async def main():
    rev = GeoLocator()
    locations = []
    # locations = [((9.964498569974612, 76.25592213325532))]
    # result = await rev.getLocationFromCoordinates(locations)
    # print(result)
    for _ in range(10):
        lat = random.uniform(-90, 90)
        lon = random.uniform(-180, 180)
        locations.append((lon, lat))
    print(await rev.getLocationFromCoordinates(locations))
    locations = []
    for _ in range(10):
        lat = random.uniform(-90, 90)
        lon = random.uniform(-180, 180)
        locations.append((lon, lat))
    print(await rev.getLocationFromCoordinates(locations))
    locations = []
    for _ in range(10):
        lat = random.uniform(-90, 90)
        lon = random.uniform(-180, 180)
        locations.append((lon, lat))

    print(await rev.getLocationFromCoordinates(locations))

    locations = []
    for _ in range(10):
        lat = random.uniform(-90, 90)
        lon = random.uniform(-180, 180)
        locations.append((lon, lat))

    print(await rev.getLocationFromCoordinates(locations))


if __name__ == '__main__':
    # asyncio.run(reverse_geocode(9.964498569974612, 76.25592213325532))
    asyncio.run(main())



import asyncio
import csv
import os
from io import StringIO

import httpx
import pandas as pd
from aiocache import cached
from aiocache.serializers import PickleSerializer

from RetroGeo.thread_type import ThreadTypeEnum
from RetroGeo.retro_geo import search

STATES_URL = f"https://raw.githubusercontent.com/dr5hn/countries-states-cities-database/master/csv/states.csv"
COUNTRIES_URL = f"https://raw.githubusercontent.com/dr5hn/countries-states-cities-database/master/csv/countries.csv"
CITIES_URL = f"https://raw.githubusercontent.com/dr5hn/countries-states-cities-database/master/csv/cities.csv"
FILENAME = "cities.csv"


class GeoLocator:
    def __init__(self):
        self.countries = {}
        self.states = {}

    async def _load_countries_states(self, url: str):
        if not self.countries or not self.states:
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                response = response.text
                csv_reader = csv.DictReader(response.splitlines())
                for row in csv_reader:
                    if row['latitude'] and row['longitude']:
                        id = int(row['id'])  # Adjust this key based on your CSV structure
                        latitude = float(row['latitude'])  # Adjust this key based on your CSV structure
                        longitude = float(row['longitude'])  # Adjust this key based on your CSV structure
                        if url == STATES_URL:
                            self.states[id] = (latitude, longitude)
                        elif url == COUNTRIES_URL:
                            self.countries[id] = (latitude, longitude)

    @cached(ttl=3600, serializer=PickleSerializer())
    async def _load_csv(self, filepath_or_buffer):
        return pd.read_csv(filepath_or_buffer)

    async def _load_cities(self):
        await asyncio.gather(*[self._load_countries_states(url) for url in [STATES_URL, COUNTRIES_URL]])
        if not os.path.exists(FILENAME):
            async with httpx.AsyncClient() as client:
                response = await client.get(CITIES_URL)
                response.raise_for_status()  # Ensure we handle errors if the request fails
                self.df = await self._load_csv(StringIO(response.text))
                df = self.df.rename(columns={
                    'latitude': 'lat',
                    'longitude': 'lon',
                    'state_id': 'admin2_id',
                    'state_name': 'admin2',
                    'country_id': 'admin1_id',
                    'country_name': 'admin1'
                })
                df = df[['lat', 'lon', 'name', 'admin1', 'admin2', 'admin1_id', 'admin2_id']]
                df['admin1_lat'] = df['admin1_id'].apply(lambda x: self.countries.get(x, (None, None))[0])
                df['admin1_lon'] = df['admin1_id'].apply(lambda x: self.countries.get(x, (None, None))[1])
                df['admin2_lat'] = df['admin2_id'].apply(lambda x: self.states.get(x, (None, None))[0])
                df['admin2_lon'] = df['admin2_id'].apply(lambda x: self.states.get(x, (None, None))[1])
                self.df = df.dropna(subset=['admin1_lat', 'admin1_lon', 'admin2_lat', 'admin2_lon'])
                self.df.to_csv(FILENAME, index=False)

    async def getLocationFromCoordinates(self, locations: list,
                                         mode: ThreadTypeEnum = ThreadTypeEnum.MULTI_THREADED.value):
        await self._load_cities()
        return search(locations, mode=mode)



""" A Parallelised Implementation of K-D Trees

Code extended from http://folk.uio.no/sturlamo/python/multiprocessing-tutorial.pdf
"""
import numpy as np
import multiprocessing as mp
import ctypes
from scipy.spatial import cKDTree

def shmem_as_nparray(shmem_array):
    """
    Function that converts a shared memory array (multiprocessing.Array) to a numpy array
    """
    return np.frombuffer(shmem_array.get_obj())

def _pquery(scheduler, data, ndata, ndim, leafsize,
            x, nx, d, i, k, eps, p, dub, ierr):
    """
    Function that parallelly queries the K-D tree based on chunks of data returned by the scheduler
    """
    try:
        _data = shmem_as_nparray(data).reshape((ndata, ndim))
        _x = shmem_as_nparray(x).reshape((nx, ndim))
        _d = shmem_as_nparray(d).reshape((nx, k))
        _i = shmem_as_nparray(i).reshape((nx, k))

        kdtree = cKDTree(_data, leafsize=leafsize)

        for s in scheduler:
            d_out, i_out = kdtree.query(_x[s, :], k=k, eps=eps, p=p, distance_upper_bound=dub)
            m_d = d_out.shape[0]
            m_i = i_out.shape[0]
            _d[s, :], _i[s, :] = d_out.reshape(m_d, 1), i_out.reshape(m_i, 1)
    except:
        ierr.value += 1

def num_cpus():
    """
    Function to get the number of CPUs / cores. This is used to determine the number of processes to spawn.
    Default (if not implemented) = 2
    """
    try:
        return mp.cpu_count()
    except NotImplementedError:
        return 2

class cKDTree_MP(cKDTree):
    """
    The parallelised cKDTree class
    """
    def __init__(self, data_list, leafsize=30):
        """ Class Instantiation
        Arguments are based on scipy.spatial.cKDTree class
        """
        data = np.array(data_list)
        n, m = data.shape
        self.shmem_data = mp.Array(ctypes.c_double, n*m)

        _data = shmem_as_nparray(self.shmem_data).reshape((n, m))
        _data[:, :] = data

        self._leafsize = leafsize
        super(cKDTree_MP, self).__init__(_data, leafsize=leafsize)

    def pquery(self, x_list, k=1, eps=0, p=2,
               distance_upper_bound=np.inf):
        """
        Function to parallelly query the K-D Tree
        """
        x = np.array(x_list)
        nx, mx = x.shape
        shmem_x = mp.Array(ctypes.c_double, nx*mx)
        shmem_d = mp.Array(ctypes.c_double, nx*k)
        shmem_i = mp.Array(ctypes.c_double, nx*k)

        _x = shmem_as_nparray(shmem_x).reshape((nx, mx))
        _d = shmem_as_nparray(shmem_d).reshape((nx, k))

        _i = shmem_as_nparray(shmem_i)
        if k != 1:
            _i = _i.reshape((nx, k))

        _x[:, :] = x

        nprocs = num_cpus()
        scheduler = Scheduler(nx, nprocs)

        ierr = mp.Value(ctypes.c_int, 0)

        query_args = (scheduler,
                      self.shmem_data, self.n, self.m, self.leafsize,
                      shmem_x, nx, shmem_d, shmem_i,
                      k, eps, p, distance_upper_bound,
                      ierr)
        pool = [mp.Process(target=_pquery, args=query_args) for _ in range(nprocs)]
        for p in pool: p.start()
        for p in pool: p.join()
        if ierr.value != 0:
            raise RuntimeError('%d errors in worker processes' % (ierr.value))

        return _d.copy(), _i.astype(int).copy()

class Scheduler:
    """
    Scheduler that returns chunks of data to be queries on the K-D Tree.
    The number of chunks is determined by the number of processes.
    """
    def __init__(self, ndata, nprocs):
        self._ndata = mp.RawValue(ctypes.c_int, ndata)
        self._start = mp.RawValue(ctypes.c_int, 0)
        self._lock = mp.Lock()
        min_chunk = ndata // nprocs
        min_chunk = ndata if min_chunk <= 2 else min_chunk
        self._chunk = min_chunk

    def __iter__(self):
        return self

    def next(self): # Python 2 support
        self._lock.acquire()
        ndata = self._ndata.value
        start = self._start.value
        chunk = self._chunk
        if ndata:
            if chunk > ndata:
                _s0 = start
                _s1 = start + ndata
                self._ndata.value = 0
            else:
                _s0 = start
                _s1 = start + chunk
                self._ndata.value = ndata - chunk
                self._start.value = start + chunk
            self._lock.release()
            return slice(_s0, _s1)
        else:
            self._lock.release()
            raise StopIteration

    def __next__(self): # Python 3 support
        self._lock.acquire()
        ndata = self._ndata.value
        start = self._start.value
        chunk = self._chunk
        if ndata:
            if chunk > ndata:
                _s0 = start
                _s1 = start + ndata
                self._ndata.value = 0
            else:
                _s0 = start
                _s1 = start + chunk
                self._ndata.value = ndata - chunk
                self._start.value = start + chunk
            self._lock.release()
            return slice(_s0, _s1)
        else:
            self._lock.release()
            raise StopIteration



""" A Fast, Offline Reverse Geocoder in Python

A Python library for offline reverse geocoding. It improves on an existing library
called reverse_geocode developed by Richard Penman.
"""
import csv
import sys
from typing import Optional

from pydantic import BaseModel, Field

from .thread_type import ThreadTypeEnum

if sys.platform == 'win32':
    # Windows C long is 32 bits, and the Python int is too large to fit inside.
    # Use the limit appropriate for a 32-bit integer as the max file size
    csv.field_size_limit(2 ** 31 - 1)
else:
    csv.field_size_limit(sys.maxsize)
from scipy.spatial import cKDTree
from . import KD_Tree
# Schema of the cities file created by this library
RG_COLUMNS = ['lat', 'lon', 'name', 'admin1', 'admin2', 'admin1_id', 'admin2_id', 'admin1_lat', 'admin1_lon',
              'admin2_lat', 'admin2_lon']
FILENAME = "cities.csv"
# WGS-84 major axis in kms
A = 6378.137

# WGS-84 eccentricity squared
E2 = 0.00669437999014


class LocationBaseModel(BaseModel):
    lat: float = Field(..., description="Latitude of the main location")
    lon: float = Field(..., description="Longitude of the main location")
    name: str = Field(..., description="Name of the location")
    admin1: str = Field(..., description="Name of the primary administrative division (e.g., country)")
    admin2: str = Field(..., description="Name of the secondary administrative division (e.g., state or province)")
    admin1_id: int = Field(..., description="ID of the primary administrative division")
    admin2_id: int = Field(..., description="ID of the secondary administrative division")
    admin1_lat: Optional[float] = Field(None, description="Latitude of the primary administrative division")
    admin1_lon: Optional[float] = Field(None, description="Longitude of the primary administrative division")
    admin2_lat: Optional[float] = Field(None, description="Latitude of the secondary administrative division")
    admin2_lon: Optional[float] = Field(None, description="Longitude of the secondary administrative division")


def singleton(cls):
    """
    Function to get single instance of the RGeocoder class
    """
    instances = {}

    def getinstance(**kwargs):
        """
        Creates a new RGeocoder instance if not created already
        """
        if cls not in instances:
            instances[cls] = cls(**kwargs)
        return instances[cls]

    return getinstance


@singleton
class RGeocoder(object):
    """
    The main reverse geocoder class
    """

    def __init__(self, mode: ThreadTypeEnum, verbose=True):
        """ Class Instantiation
        Args:
        mode (int): Library supports the following two modes:
                    - 1 = Single-threaded K-D Tree
                    - 2 = Multi-threaded K-D Tree (Default)
        verbose (bool): For verbose output, set to True
        stream (io.StringIO): An in-memory stream of a custom data source
        """
        self.mode = mode
        self.verbose = verbose
        coordinates, self.locations = self.load()
        if mode == ThreadTypeEnum.SINGLE_THREADED:  # Single-process
            self.tree = cKDTree(coordinates)
        else:  # Multi-process
            self.tree = KD_Tree.cKDTree_MP(coordinates)

    def query(self, coordinates):
        """
        Function to query the K-D tree to find the nearest city
        Args:
        coordinates (list): List of tuple coordinates, i.e. [(latitude, longitude)]
        """
        if self.mode == 1:
            _, indices = self.tree.query(coordinates, k=1)
        else:
            _, indices = self.tree.pquery(coordinates, k=1)
        len_ = len(self.locations)
        return [self.locations[index] for index in indices if index<len_]

    def load(self):
        """
        Function that loads a custom data source
        Args:
        stream (io.StringIO): An in-memory stream of a custom data source.
                              The format of the stream must be a comma-separated file
                              with header containing the columns defined in RG_COLUMNS.
        """
        with open(FILENAME, mode='r', newline='') as file:
            stream_reader = csv.DictReader(file)
            header = stream_reader.fieldnames
            if header != RG_COLUMNS:
                raise csv.Error('Input must be a comma-separated file with header containing ' + \
                                'the following columns - %s. For more help, visit: ' % (','.join(RG_COLUMNS)) + \
                                'https://github.com/thampiman/reverse-geocoder')

            # Load all the coordinates and locations
            geo_coords, locations = [], []
            for row in stream_reader:
                geo_coords.append((row['lat'], row['lon']))
                locations.append(row)
            return geo_coords, locations


def search(geo_coords, mode, verbose=False):
    """
    Function to query for a list of coordinates
    """
    if not isinstance(geo_coords, tuple) and not isinstance(geo_coords, list):
        raise TypeError('Expecting a tuple or a tuple/list of tuples')
    elif not isinstance(geo_coords[0], tuple):
        geo_coords = [geo_coords]
    _rg = RGeocoder(mode=mode, verbose=verbose)
    return dict(zip(geo_coords,[LocationBaseModel(**result) for result in _rg.query(geo_coords)]))



from setuptools import setup, find_packages

VERSION = '0.0.3'
DESCRIPTION = 'RetroGeo:  A simple offline Reverse Geocoder Library'
LONG_DESCRIPTION = '''
Our reverse geocoding library is a robust Python tool designed to translate latitude and longitude coordinates into detailed addresses, entirely offline. It processes geographical coordinates, such as latitude and longitude, and returns a full address corresponding to the location, including country, state, city. The library is optimized for performance and accuracy, leveraging pre-downloaded datasets of geographical locations, eliminating the need for online queries or API calls.

This library is especially useful for applications where internet connectivity is limited or unavailable, or where privacy concerns require offline processing. It supports large-scale reverse geocoding operations by efficiently handling massive datasets, making it ideal for geospatial analysis, mobile applications, and data processing pipelines that require rapid location lookup without external dependencies.

With built-in data compression and storage optimizations, our reverse geocoding library minimizes disk space while maintaining high accuracy, offering developers a reliable solution for offline geocoding tasks.'''

# Setting up
setup(
    name="RetroGeo",
    version=VERSION,
    author="SOORAJ TS",
    author_email="sjts007@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
        'aiocache==0.12.2',
        'annotated-types==0.7.0',
        'anyio==4.4.0',
        'certifi==2024.8.30',
        'charset-normalizer==3.3.2',
        'docutils==0.21.2',
        'h11==0.14.0',
        'httpcore==1.0.5',
        'httpx==0.27.2',
        'idna==3.8',
        'importlib_metadata==8.4.0',
        'jaraco.classes==3.4.0',
        'jaraco.context==6.0.1',
        'jaraco.functools==4.0.2',
        'keyring==25.3.0',
        'markdown-it-py==3.0.0',
        'mdurl==0.1.2',
        'more-itertools==10.5.0',
        'nh3==0.2.18',
        'numpy==2.1.1',
        'pandas==2.2.2',
        'pkginfo==1.10.0',
        'pydantic==2.9.0',
        'pydantic_core==2.23.2',
        'Pygments==2.18.0',
        'python-dateutil==2.9.0.post0',
        'pytz==2024.1',
        'readme_renderer==44.0',
        'requests==2.32.3',
        'requests-toolbelt==1.0.0',
        'rfc3986==2.0.0',
        'rich==13.8.0',
        'scipy==1.14.1',
        'setuptools==74.1.2',
        'six==1.16.0',
        'sniffio==1.3.1',
        'typing_extensions==4.12.2',
        'tzdata==2024.1',
        'urllib3==2.2.2',
        'zipp==3.20.1',
    ],
    # needs to be installed along with your package. Eg: 'caer'

    keywords=['reverse geocoding', 'geolocation', 'offline', 'python'],
    classifiers=[
        "Development Status :: 4 - Beta",  # Indicating the package is more stable than Alpha
        "Intended Audience :: Developers",  # Broader audience targeting developers
        "Topic :: Software Development :: Libraries :: Python Modules",  # Relevant to Python modules
        "License :: OSI Approved :: MIT License",  # Update based on your license type
        "Programming Language :: Python :: 3",  # General support for Python 3
        "Operating System :: OS Independent",  # If your package works across multiple platforms
    ]
)
